"""

Python Desktop Server - Wiki Editor

Copyright (c) 2002, Georg Bauer <gb@murphy.bofh.ms>
Copyright (c) 2002, Garth T Kidd <garth@deadlybloodyserious.com>

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to 
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 
the Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

# $Id: WikiTool.py,v 1.13 2003/10/23 14:49:35 gb Exp $

import PyDS.Tool
import metakit
import time
import string
import sys
import cgi
import re
import os.path
    
def _munge(s): 
	"Elminiate non-alphanumeric and make lowercase."
	t = re.sub('\W', '', s)
	return t.lower()

def _fullSplit(path): 
	"split a path to a tuple, with '' as the first item"
	p, b = os.path.split(path)
	if not p: 
		return (b,)
	else: 
		return _fullSplit(p) + (b,)

class WikiTool(PyDS.Tool.StandardTool):

	def _recachepostings(self):
		raise NotImplementedError # can't see a need for it

	def _initdb(self):
		try:
			self._acquire()
			self.nodes = self.db.getas(
				"nodes[id:S,title:S,text:S,pubdate:S,pubtime:F,categoryid:S,baseid:S]"
			).ordered(1)
			self.categories = self.db.getas(
				"categories[id:S]"
			).ordered(1)
			self._commit()
		finally: self._release()

	
	def _backup(self):
		l = self._backupTable('nodes')
		return l
			
	def _initopts(self):
		self.rss = 1
		self.showCalendar =0 
		if self._getNode('index') is None: 
			# If I add a preference for rendering Wiki pages to the main site 
			# rather than a tool-specific subdirectory, I'll need to suppress 
			# automatic creation of this file. 
			self.addNode('index', 
			             _("Wiki Index"), 
				     _("Edit this page to begin your Wiki"))
		if self._getNode('404') is None: 
			self.addNode('404', 
			             _("Unresolved Target"), 
				     _("Not-yet-resolved links in the Wiki are pointed here."))
		navigation = self.getToolByNamespace('navigation')
		navigation._addHook('select', 'wiki',
			_('Link to a wiki page'),
			self._wikiNavHook)
		
	def _wikiNavHook(self, qual, value, preset=None):
		if qual == 'html':
			html = '<select name="%s">' % value
			for node in self.getNodes():
				selected = ''
				if node['id'] == preset:
					selected = ' selected'
				html += '<option value="%s"%s>%s' % (
					node['id'],
					selected,
					node['title']
				)
			html += '</select>'
			return html
		elif qual == 'url':
			return self.getCloudUrlForNode(value)
		else:
			raise ValueError('%s is not a valid qualifier' % qual)
	
	def _initthread(self, server):
		while server.isRunning():
			time.sleep(10)
			if self.queue.length():
				while self.queue.length():
					(jobf, key, data) = self.queue.shift()
					jobf(key, data)
				# since we had something to render, we should
				# now render the RSS feed to get a current one
				req = self._getDummyRequest()
				xml = req.renderPage('RSSFeedRendering', category='')
				self.storeXML(xml, category='')
				self.logVerbose(_('work done, going to sleep again'))
	
	def _status(self):
		try:
			self._acquire()
			html = PyDS.Tool.StandardTool._status(self)
			html += '<ul>'
			html += _('<li>There are %d wiki nodes in the database</li>') % len(self.nodes)
			html += _('<li>There are %d items in the render queue</li>') % self.queue.length()
			html += self._threadstatus()
			html += '</ul>'
			return html
		finally: self._release()
	
	def _search(self, query):
		return [] # not implemented yet

	# this method returns an URL to the cloud in the context of the current tool
	# or the given tool. Weblog patches the URL if radio compatibility is requested.
	def getStaticUrl(self, page=None, tool=None, year=None, month=None, day=None, category='', base=None, flat=0):
		if filter(None, (year, month, day, category)): 
			pass # aah, ignore it for the time being 
			# raise NotImplementedError, _("argument not supported by Wiki tool")
		if page is None: 
			pass # aah, ignore it for the time being
			# raise ValueError, _("page argument required by Wiki tool")
		url = PyDS.Tool.StandardTool.getStaticUrl(self, page, tool, year, month, day, category, base, flat)
		return url

	# This method returns the URL for a node. Depending on context it
	# returns a desktop URL or a cloud URL.
	def getCloudUrlForNode(self, id): 
		if _flet.desktop:
			return self.getUrl('show_html', id=str(id))
		else:
			node = self.getNode(id)
			if node:
				return self.getStaticUrl(page='%s.html' % id)
			else:
				return self.getStaticUrl(page='404.html')
		
	# --------------------------------------------------------------------
	# NEW STRAY LINK RESOLUTION METHODS
	# --------------------------------------------------------------------

	# Find items potentially matching the link text, and return their unique IDs. 
	def getUniqueIdsMatchingLinkText(self, linkText): 
		"""Find items potentially matching `linkText`. 
		Return a list of string IDs uniquely identifying those items."""
		if not linkText: 
			return # never match empty link text

		# precalculate more permissive versions of the link text
		m = _munge(linkText)
		lt = linkText.lower()

		# precalculate permissive versions of context-dependent variables
		conid = _flet.get('uniqueId', '').lower() # uniqueId
		concatid = _flet.get('categoryid', '').lower() # categoryid
		
		# define filters, each of which depends on local variables defined 
		# previously -- a lot easier than passing lots of arguments, no?
		orderedFilters = (
			# id match to link text
			lambda: lid == lt, 
			# id match on munged text
			lambda: lid == m, 
			# subnode of current (according to context) node
			lambda: conid and catid == conid and baseid == m, 
			# peer node of current (according to context) node
			lambda: concatid and catid == concatid and baseid == m, 
			# baseid match
			lambda: baseid == m, 
			# title match
			lambda: m == _munge(node.title()),
			)

		# prepare filter results dictionary
		filterResults = {}
		for filter in orderedFilters: 
			filterResults[filter] = []

		# run each node through the filters
		# if CPU becomes a bigger worry than memory, we can cache all these
		try: 
			self._acquire()
			for node in self.nodes: 
				# precalculate more permissive version of node attributes
				nid = node.id
				lid = nid.lower()
				catid = node.categoryid.lower()
				baseid = node.baseid.lower()
				# execute the filters
				for filter in orderedFilters: 
					try: 
						if filter(): 
							filterResults[filter].append(nid)
					except: 
						pass # not worried about filter failure
		finally:
			self._release()

		# gather results into a single list, preserving order
		allResults = []
		inAllResults = {}
		for filter in orderedFilters: 
			for result in filterResults[filter]: 
				if inAllResults.has_key(result): 
					continue
				else: 
					inAllResults[result] = 1
					allResults.append(result)
		return allResults

	# Return the upload target for a given item in the current context. 
	def getCloudLinkDetailsForUniqueId(self, uniqueId): 
		"""Return the upload target for `uniqueId` in the current context.
		Raise KeyError if `uniqueId` cannot be found."""
		node = self.getNode(uniqueId)
		if node:
			return ('wiki', self.getStaticUrl(page='%s.html' % uniqueId), node.get('title',''))
		else: 
			raise KeyError, uniqueId

	# Return a URL for an edit page for a given item.
	def getDesktopDisplayLinkDetailsForUniqueId(self, uniqueId): 
		"""Return the edit URL for `uniqueId`."""
		node = self.getNode(uniqueId)
		if node:
			return ('wiki', self.getUrl('show_html', id=uniqueId), node.get('title',''))
		else: 
			raise KeyError, uniqueId

	# Return a link to a creation page for an item, based on link text. 
	def getDesktopCreateLinkDetailsForProposedLinkText(self, linkText): 
		"""Return the edit URL for a non-existent item, based on `linkText`."""
		return ('undefinednode', self.getUrl('edit_html', id=_munge(linkText)))

	# Render an item by its unique ID
	def renderItemByUniqueId(self, uniqueId):
		return self.renderNode(uniqueId)

	# --------------------------------------------------------------------
		
	# this method returns the title of a page for rendering. This does a
	# lookup of data in the category or in the preferences and adds
	# date specifiers. The context passed in is the context that you
	# get in your template.
	def getPageTitle(self, context):
		if context.has_key('title'): 
			return context['title']
		log('WikiTool.getPageTitle failed to find title; context.keys==[%s]', 
		    ', '.join(map(repr, context.keys())))
		title = PyDS.Tool.StandardTool.getPageTitle(self, context)
		if context.get('category',''):
			cat = self.getCategory(context['category'])
			if cat:
				title = cat['title']
		if context.get('year',0) and context.get('month',0) and context.get('day',0):
			title += ' %d.%d.%d' % (context['day'], context['month'], context['year'])
		elif context.get('year',0) and context.get('month',0):
			title += ' %d/%d' % (context['year'], context['month'])
		return title
		
	# this method returns the description of a page for rendering. It's
	# a sister to the getPageTitle method.
	def getPageDescription(self, context):
		desc = PyDS.Tool.StandardTool.getPageDescription(self, context)
		if context.get('category',''):
			cat = self.getCategory(context['category'])
			if cat:
				desc = cat['desc']
		return desc
		
	def index_html(self, req):
		macros = self.getToolByNamespace('macros')
		html = [_("<h3>Your Wiki Nodes</h3>")]
		append = html.append
		editvars = {
			'edit_html': self.getUrl('edit_html'),
			'editlabel': _("New")
		}
		append('<form action="%(edit_html)s" method="GET">' % editvars)
		append('<input type="submit" value="%(editlabel)s" accesskey="n">' % editvars)
		try:
			self._acquire()
			hashedNodes = {}
			for node in self.nodes: 
				nids = _fullSplit(node.id)
				current = hashedNodes
				for id in nids: 
					idnode = current.setdefault(id, [None, {}])
					current = idnode[1]
				idnode[0] = node
			def listRenderer(hashedNodes): 
				keys = hashedNodes.keys()
				keys.sort()
				html.append('<ul>')
				for k in keys: 
					html.append('<li>')
					node, hashedChildren = hashedNodes[k]
					if node is not None: 
						nodevars = {
							'url': self.getUrl('show_html', id=str(node.id)),
							'title': node.title,
							'id': k,
							'lastmod': macros.Today(node.pubtime, withtime=1),
						}
						html.append('<a href="%(url)s" title="%(lastmod)s">%(id)s</a>: %(title)s' % nodevars)
					else: 
						html.append('%s>' % k)
					if hashedChildren: 
						listRenderer(hashedChildren)
					html.append('</li>')
				html.append('</ul>')
			listRenderer(hashedNodes)
		finally: self._release()
		html.append('</form>')
		req.setLocalValue('title', _('Your Wiki Nodes'))
		return req.renderPage('BaseTemplate', body='\n'.join(html))

	def show_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		node = self.getNode(id)
		# used to be wrapped in 
		# _flet.begin(categoryid=node.get('categoryid',''), id=id)
		rendered = req.renderText(
			node.get('text',_('This node has no text.')), 
			structured = 1,
			uniqueId = id, 
			homeTool = self)
		# breadcrumbs
		nids = _fullSplit(id)
		sofar = ''
		crumbs = ['<strong><a href="%s">%s</a></strong>' % (self.getUrl('index_html'), _('Wiki'))]
		for nid in nids: 
			if sofar: 
				sofar = sofar + '/' + nid
			else: 
				sofar = nid
			if self._getNode(sofar) is None: 
				cssclass = 'undefinednode'
			else: 
				cssclass = 'node'
			crumbs.append('<a class="%s" href="%s">%s</a>' % (
				cssclass, self.getUrl('show_html', id=sofar), nid
			))
		html = """\
<p>%(crumbs)s</p>
<h2>%(title)s</h2>
%(rendered)s
<form action="%(mux_redir)s" method="POST">
<input type="hidden" name="id" value="%(id)s">
<input type="submit" name="WikiEdit" value="%(editlabel)s" accesskey="e">
<input type="submit" name="WikiDelete" value="%(deletelabel)s">
<input type="submit" name="WikiView" value="%(viewlabel)s" accesskey="v">
</form>
<p align="right" class="smallfont">%(lastlabel)s %(pubtimestr)s</p>
<p align="right"><a href="%(toolurl)s" accesskey="i">%(index)s</a></p>""" % {
			'title': node.get('title',_('Set your page title here')), 
			'crumbs': '&nbsp;/&nbsp;'.join(crumbs), 
			'mux_redir': self.getUrl('mux_redir'),
			'id': str(id), 
			'editlabel': _("Edit"), 
			'viewlabel': _("View"), 
			'deletelabel': _("Delete"), 
			'rendered': rendered, 
			'lastlabel': _("last change"), 
			'pubtimestr': macros.Today(node.get('pubtime',time.time()), withtime=1), 
			'toolurl': self.getUrl(), 
			'viewurl': self.getStaticUrl(page='%s.html' % id),
			'index': _("Wiki index"),
		}
		req.setLocalValue('title', node.get('title',_('Set your page title here')))
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Show: %s" % node.get('title', '(no title)'))
		
	def mux_redir(self, req): 
		id = req.getFirstValue('id')
		if not id: 
			id = ''
		if req.getFirstValue('WikiEdit'): 
			return self.getUrl('edit_html', id=str(id))
		if req.getFirstValue('WikiDelete'): 
			return self.getUrl('delete_html', id=str(id))
		if req.getFirstValue('WikiView'): 
			return self.getStaticUrl(page='%s.html'%id)
		return self.getUrl('show_html', id=str(id))

	def delete_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		node = self.getNode(id)
		html = """\
%(rendered)s
<form action="%(delete_redir)s" method="POST">
<input type="hidden" name="id" value="%(id)s">
<input type="submit" name="WikiDeleteConfirm" value="%(confirmlabel)s" accesskey="d">
<input type="submit" name="WikiDeleteCancel" value="%(cancellabel)s" accesskey="c">
</form>
<p align="right" class="smallfont">%(lastlabel)s %(pubtimestr)s</p>""" % {
			'title': node['title'], 
			'show_html': self.getUrl('show_html', id=str(id)), 
			'delete_redir': self.getUrl('delete_redir', id=str(id)), 
			'id': str(id), 
			'cancellabel': _("Cancel"), 
			'confirmlabel': _("Confirm Deletion"), 
			'rendered': req.renderText(
				node['text'], 
				structured = 1, 
				#categoryid=node['categoryid'], 
				#id=id,
				uniqueId = id, 
				homeTool = self),
			'lastlabel': _("last change"), 
			'pubtimestr': macros.Today(node['pubtime'], withtime=1), 
		}
		req.setLocalValue('title', node['title'])
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Delete: %s" % node.get('title', '(no title)'))

	def edit_html(self, req):
		macros = self.getToolByNamespace('macros')
		id = req.getFirstValue('id')
		if req.getFirstValue('WikiView'): 
			return self.getStaticUrl(page='%s.html' % id)
		if not id:
			id = ''
		node = self.getNode(id)
		values = {} 
		for field in ['id', 'title', 'text']: 
			value = ''
			if id: # defaults
				value = id
			if node:
				value = cgi.escape(node[field])
			values[field+'value'] = value
		for field in ['Id', 'Title', 'Text', 'delete', 'save']: 
			values[field.lower()+'label'] = _(field)
		values['edit_redir'] = self.getUrl('edit_redir')
		html = """
<form action="%(edit_redir)s" method="POST">
<table border=0>
<tr>
  <th align="right">%(idlabel)s</th>
  <td><input name="oldid" value="%(idvalue)s" type="hidden">
      <input name="id" value="%(idvalue)s" size=50></td>
</tr><tr>
  <th align="right">%(titlelabel)s</th>
  <td><input name="title" value="%(titlevalue)s" size=50></td>
</tr><tr>
  <th valign="top" align="right">%(textlabel)s</th>
  <td><textarea nowrap name="text" rows=15 cols=60>%(textvalue)s</textarea></td>
</tr><tr>
  <td colspan=2 align="right">
    <input type=submit name="WikiDelete" value="%(deletelabel)s">
    <input type=submit name="WikiSave" accesskey="s" value="%(savelabel)s">
  </td>
</tr></table></form>
""" % values
		html += '<p align="right"><a href="%s">%s</a>' % (
			self.getUrl(),
			_('choose a different node')
		)
		req.setLocalValue('title', _('Edit your node below'))
		return req.renderPage('BaseTemplate', body=html, title="Wiki: Edit: %s" % node.get('title', '(no title)'))

	def delete_redir(self, req):
		id = req.getFirstValue('id')
		if req.getFirstValue('WikiDeleteCancel'): 
			return self.getUrl('show_html', id=str(id))
		self.deleteNode(id)
		return self.getUrl()

	def edit_redir(self, req):
		id, oldid, title, text, desc = map(req.getFirstValue, 
			('id', 'oldid', 'title', 'text', 'desc'))
		if req.getFirstValue('WikiDelete'):
			return self.getUrl('delete_html', id=str(id))
		else:
			if not id:
				id = ''
			if oldid and text and oldid != id: 
				self.deleteNode(oldid, noRender=1) # the addNode will get it
			if id and text:
				id = self.addNode(id, title, text)
				return self.getUrl('show_html', id=str(id))
			else:
				if id:
					return self.getUrl('edit_html', id=str(id), errmsg=_('You need to give id and text!'))
				else:
					return self.getUrl('edit_html', errmsg=_('You need to give id and text!'))

	# this method deletes an existing node and drops the rendered output
	def deleteNode(self, id, noRender=0):
		try:
			self._acquire()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				self.nodes.delete(idx)
				self._commit()
				self.queue.append(
					self._deleteRenderedNode,
					id,
					None
				)
			return found
		finally: self._release()
		meshTool = self.getToolByNamespace('mesh')
		meshTool.delItem(id, self)
		if not noRender:
			meshTool.renderItemsWithUnresolvedLinks()

	def _deleteRenderedNode(self, id, dummy):
		self.deleteFile('%s.html' % id)

	# This method delivers the description of your preferences panels
	# to hook up into the preferences list.
	def getPreferencePanels(self):
		return [] # not implemented
			
	# Internal getNode; returns the row, not a dict.
	def _getNode(self, id): 
		try:
			self._acquire()
			(res,found) = self.nodes.locate({'id':id})
			if found:
				return self.nodes[res]
			else:
				return None
		finally: self._release()

	# get a posting by ID
	def getNode(self, id):
		node = self._getNode(id)
		if node is None: 
			return {}
		else: 
			return {
				'id': node.id,
				'title': node.title,
				'text': node.text,
				'pubtime': node.pubtime,
				'categoryid': node.categoryid, 
				'baseid': node.baseid, 
			}
	
	# get a list of all nodes (without their text!)
	def getNodes(self):
		try:
			self._acquire()
			liste = []
			for node in self.nodes:
				liste.append({
					'id': node.id,
					'title': node.title,
					'pubtime': node.pubtime,
					'categoryid': node.categoryid,
					'baseid': node.baseid
				})
			return liste
		finally: self._release()

	def addNode(self, id, title, text): 
		categoryid, baseid = os.path.split(id)
		macros = self.getToolByNamespace('macros')
		hook = ''
		try:
			self._acquire()
			pubtime = time.time()
			(idx, found) = self.nodes.locate({'id':id})
			if found:
				node = self.nodes[idx]
				if node.title != title or node.text != text or node.categoryid != categoryid or node.baseid != baseid:
					# TODO: if the ID is unchanged, category and baseid must also be unchanged, right? 
					node.title = title
					node.text = text
					node.pubtime = pubtime
					node.categoryid = categoryid
					node.baseid = baseid
					self._commit()
					self.renderNode(id) 
			else: # NEW!
				self.nodes.append({
					'id': id,
					'title': title,
					'text': text,
					'pubtime': pubtime,
					'baseid': baseid,
					'categoryid': categoryid
				})
				self._commit()
				self.renderNode(id)
				meshTool = self.getToolByNamespace('mesh')
				meshTool.renderItemsWithUnresolvedLinks()
			return id
		finally:
			self._release()
			if hook and hasattr(macros, hook):
				getattr(macros, hook)(id)

	def _rerender(self):
		try:
			self._acquire()
			count = 0
			for node in self.nodes:
				self.renderNode(node.id)
				count += 1
			return count
		finally: self._release()

	# this method renders a single page in the background.
	def renderNode(self, id):
		self.queue.append(self.renderNodeOnline, (id,), None)
		return self.queue.length()

	# this method renders a single page, but does it online (not queued)
	def renderNodeOnline(self, key, data):
		(id,) = key
		node = self.getNode(id=id)
		pagename = '%s.html' % id
		if os.sep=='\\': 
			pagename = re.sub('/',r'\\',pagename)
		req = self._getDummyRequest()
		toolname = self.name # consider ''
		body = req.renderText(
			node['text'], 
			structured = 1,
			uniqueId = id,
			homeTool = self)
		html = req.renderPage('BaseRendering', body=body, category='', title=node.get('title', '(no title)'))
		fn = self.storeHTML(html, page=pagename, toolname=toolname)
		return fn

	# this method returns the title of the RSS feed
	def getRSSTitle(self, category=''):
		try:
			self._acquire()
			prefs = self.getPreferencesTool()
			title = prefs.get('private', 'title')
			if title:
				title += ': '
				title += _('Wiki')
			else:
				title = _('An Unconfigured Python Desktop Server: Wiki')
			return title
		finally: self._release()
	
	# this method returns the description of the RSS feed
	def getRSSDescription(self, category=''):
		try:
			self._acquire()
			prefs = self.getPreferencesTool()
			desc = prefs.get('private', 'description')
			return desc
		finally: self._release()
	
	# this method returns the stories as an item list, orderd by time
	def getRSSItems(self, category=''):
		try:
			self._acquire()
			macros = self.getToolByNamespace('macros')
			items = []
			for node in self.nodes:
				desc = node.text
				if len(desc) > 250:
					desc = desc[:250] + ' ...'
				item = {
					'title': macros.quoteXML(node.title),
					'link': self.getCloudUrlForNode(node.id),
					'guid': self.getCloudUrlForNode(node.id),
					'pubdate': macros.TodayFull(now=node.pubtime),
					'pubtime': node.pubtime,
					'source': '',
					'sourceurl': '',
					'categories': [],
					'comments': '',
					'trackback': '',
					'description': macros.quoteXML(desc)
				}
				items.append(item)
			items.sort(lambda a,b: -1*cmp(a['pubtime'],b['pubtime']))
			if len(items) > 20:
				items = items[:21]
			for item in items:
				del item['pubtime']
			return items
		finally: self._release()

	# this method returns hours without stories
	def getRSSSkipHours(self, category=''):
		try:
			self._acquire()
			hours = {}
			for node in self.nodes:
				hours[time.gmtime(node.pubtime)[3]] = 1
		finally: self._release()
		nonhours = []
		for hour in range(24):
			if not(hours.has_key(hour)):
				nonhours.append(hour)
		return nonhours


__desc__ = _("This tool allows to edit wiki postings.")

PyDS.Tool.registerTool(WikiTool('wiki', _('Wiki'), __desc__, 150))

